#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "ruleset.h"
#include "mainwin.h"
#include "support.h"

#define RULESET_TEMPFILE "/tmp/ruleset.fwx"

gchar* ruleset_file = NULL;
gboolean ruleset_modified = FALSE;
static GList *ruleset = NULL;

static pid_t child_pid;

static GtkListStore *ruleset_list_store = NULL;
enum { COLUMN_ENABLED, COLUMN_DESCRIPTION, N_COLUMNS };


void ruleset_free() {
    while (ruleset) {
        rule_free(ruleset->data);
        ruleset = g_list_remove(ruleset, ruleset->data);
    }


    ruleset_modified = FALSE;
}


int _ruleset_load(FILE *f) {
    return 0;
}

int _ruleset_save(FILE *f) {
    return 0;
}


int ruleset_load(gchar *fn) {
    int r;
    FILE*f;
    
    if (!(f = fopen(fn, "r")))
        return -1;

    r = _ruleset_load(f);
    fclose(f);
    return r;
}

int ruleset_save(gchar *fn) {
    int r;
    FILE*f;
    if (!(f = fopen(fn, "w")))
        return -1;

    r = _ruleset_load(f);
    fclose(f);
    return r;
}


int ruleset_commit() {
    char *fn = ruleset_file;
    char *argv[3];
    GError *e;
    
    if (!fn || ruleset_modified) {
        int fd;
        FILE *f;
        for (;;) {
            if ((fd = open(fn = RULESET_TEMPFILE, O_EXCL|O_WRONLY)) < 0 && errno != EEXIST) {
                fprintf(stderr, "Could not create temporary file: %s\n", strerror(errno));
                return -1;
            }

            if (fd >= 0)
                break;
            
            if (unlink(fn)  < 0) {
                fprintf(stderr, "Could not unlink temporary file: %s\n", strerror(errno));
                return -1;
            }
        }
        
        if (!(f = fdopen(fd, "w"))) {
            close(fd);
            return -1;
        }

        if (_ruleset_save(f) < 0)
            return -1;

        fclose(f);
    }


    g_assert(fn);

    argv[0] = "../helper/install-firewall";
    argv[1] = fn;
    argv[2] = NULL;

    if (!g_spawn_async_with_pipes(NULL, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &child_pid, NULL, NULL, NULL, &e)) {
        fprintf(stderr, "Could not run install-firewall: %s\n", e->message);
        g_error_free(e);
        return -1;
    }

    return 0;
        
}

int ruleset_append_rule(rule_t *rule) {
    g_assert(rule);
    
    if (g_list_find(ruleset, rule))
        return -1;
    
    ruleset = g_list_append(ruleset, rule);

    gtk_list_store_append(ruleset_list_store, &rule->iter);
    gtk_list_store_set(ruleset_list_store, &rule->iter,
                       COLUMN_ENABLED, rule->enabled,
                       COLUMN_DESCRIPTION, rule->description, 
                       -1);

    rule->realized = TRUE;
    
    return 0;
}

int ruleset_update_rule(rule_t *rule) {
    g_assert(rule && rule->realized);

    gtk_list_store_set(ruleset_list_store, &rule->iter,
                       COLUMN_ENABLED, rule->enabled,
                       COLUMN_DESCRIPTION, rule->description, 
                       -1);

    return 0;
}


int ruleset_remove_rule(rule_t *rule) {
    g_assert(rule);
    
    if (!g_list_find(ruleset, rule))
        return -1;

    gtk_list_store_remove(ruleset_list_store, &rule->iter);
    
    ruleset = g_list_remove(ruleset, rule);
    return 0;
}


void ruleset_widget_init() {
    GtkTreeView *tv = GTK_TREE_VIEW(lookup_widget(get_main_window(), "ruleset_view"));

    ruleset_list_store = gtk_list_store_new(N_COLUMNS, G_TYPE_BOOLEAN, G_TYPE_STRING);
    gtk_tree_view_set_model(tv, GTK_TREE_MODEL(ruleset_list_store));

    gtk_tree_view_append_column(tv, gtk_tree_view_column_new_with_attributes("Enabled", gtk_cell_renderer_toggle_new(), "active", COLUMN_ENABLED, NULL));
    gtk_tree_view_append_column(tv, gtk_tree_view_column_new_with_attributes("Description", gtk_cell_renderer_text_new(), "text", COLUMN_DESCRIPTION, NULL));
}

rule_t* rule_new() {
    rule_t* rule = g_new0(rule_t, 1);

    strcpy(rule->device_in, "eth0");
    strcpy(rule->device_out, "eth0");
    rule->direction = DIR_OUTGOING;

    rule->id = 0;
    
    return rule;
}

rule_t* rule_new_from_conn_info(conn_info_t *ci) {
    rule_t *rule = rule_new();
    g_assert(ci);

    rule->enabled = TRUE;
    rule->realized = FALSE;
    strncpy(rule->description, ci->type_string, sizeof(rule->description));
    strncpy(rule->device_in, ci->device_in, IFNAMSIZ);
    strncpy(rule->device_out, ci->device_out, IFNAMSIZ);

    rule->protocol = ci->protocol;
    rule->port = ci->port;
    rule->icmp_type = ci->icmp_type;
    rule->direction = ci->direction;

    rule->src_ip_address = ci->src_ip_address;
    rule->dst_ip_address = ci->dst_ip_address;
    
    rule->src_netmask_bits = 32;
    rule->dst_netmask_bits = 32;
    
    return rule;
}
